#######################################################
###  __init__.py
###  This is the fundamental python module.  I import 
###  everything into this namespace to shorten the 
###  calling tree when this module is used in other
###  scripts.  This also contains the version.
#######################################################

from subnetfinder.subnetfinder import *
__version__='1.2.1'


#######################################################
###  __main__.py
###  This script is called when the module is executed
###  from the command line, via python -m <mod>.
###  Therefore, this will rationalize the options 
###  provided from the CLI and pass them to the module
###  method.
#######################################################

import ipaddress
import argparse
import sys
import subnetfinder.subnetfinder as subnetfinder
from subnetfinder.__init__ import __version__

def main():
    '''
    This method will rationalize the input so that the "ProcessSubnets" method can be called
    as a module from other scripts.
    '''
    
    ###################################
    ###  This code is only used if this
    ###  module is called from the CLI
    ###  directly rather than being 
    ###  called as a module.
    ###################################

    parser=argparse.ArgumentParser(description='Find unused subnets from a provided list of used subnets.',
    prog='subnetfinder',
        epilog='''
        Use this program to find unused blocks of IPs, or
        blocks used a limited number of times, with 
        prefix length pl within the subnet sn, based on 
        the IPs or blocks supplied.''')
    parser.add_argument('-s','--supernet',metavar='sn', type=ipaddress.ip_network, dest='sn',
        help='The Super Net in which to find open IP blocks, such as 10.0.0.0/8.',
        required=True)
    parser.add_argument('-p','--prefix',metavar='pl',type=int,dest='pl',
        help='The maximum prefix length to consider, such as 16.  This must be longer than the supernet prefix length.',
        required=True)
    parser.add_argument('-u','--uniquein',metavar='N',type=int,dest='uniquein',default=0,
        help='Specify the maximum number of groups that can contain a prefix.  Can be used to find\
            prefixes unique to two or more lists.  Groups must have headings of "group: <groupname>".')
    parser.add_argument(nargs='?',type=argparse.FileType('r'),default=sys.stdin,dest='FILE',
        help='If present, the file name of the source list with one prefix per line.  Default is from STDIN.')
    parser.add_argument('--summarize',action='store_true',
        help='Summarize contiguous subnets into supernets if possible.')
    parser.add_argument('--version',action='version',version='%(prog)s version '+ __version__)
    parser.add_argument('-v','--verbose',action='count',default=0,
        help='Be more verbose with output. Verbose levels 1 to 3 are -v to -vvv.')
    parser.add_argument('--debug',action='store_true',
        help='Display additional output for lines.')
    parser.add_argument('--test',action='store_true',
        help='Used only for unit testing')
    parsed=parser.parse_args()

    if (parsed.sn.prefixlen>=parsed.pl):
        parser.error('The prefix length of the Super Net supplied must be shorter than the target prefix length. You gave me a Super Net prefix of ('+
        str(parsed.sn.prefixlen)+') and a target prefix of ('+str(parsed.pl)+').')
    
    file=parsed.FILE
    supernet=parsed.sn
    prefixlen=parsed.pl
    debug=parsed.debug
    summarize=parsed.summarize
    uniquein=parsed.uniquein
    verbose=parsed.verbose
    if debug:
        print("Options Provided:")
        print(" Supernet: %s"%(supernet))
        print(" Prefix Len: %d"%(prefixlen))
        print(" Unique In: %d"%(uniquein))
        print(" Debug? %s"%(debug))
        print(" Summarize? %s"%(summarize))
        print(" Verbose Level: %d"%(verbose))
    usedlist=file.readlines()
    if parsed.test:
        return parsed
    else:
        availablelist=subnetfinder.ProcessSubnets(supernet,usedlist,prefixlen,showdebug=debug,summarize=summarize,screenout=True,uniquein=uniquein)
    
    print("Available Subnets:")
    for x in [str(x) for x in availablelist ]:
        print("%s"%(x))

    if (uniquein > 0) & (verbose>=1):
        print("Subnets which occur in no more than "+ str(uniquein) +" group(s)")
        for x in [str(x) for x in subnetfinder.uniquelist]:
            print("%s"%(x))

    return

if __name__=='__main__':
    main()



import pathlib
from setuptools import setup, find_packages
from subnetfinder.__init__ import __version__

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="subnetfinder",
    version=__version__,
    python_requires='>=3.7',
    description="Find available subnets within a given block of IPs",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://sourceforge.net/projects/available-subnet-finder/",
    author="Skylight 2000",
    #author_email="",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
    ],
    #packages=["subnetfinder"],
    scripts=[],
    packages=find_packages(include=['subnetfinder','subnetfinder.*'],exclude=('tests',)),
    include_package_data=True,
    #install_requires=["math", "sys", "ipaddress"],
    entry_points={
        'console_scripts': [
            'subnetfinder=subnetfinder.__main__:main'
        ],
        'subnetfinder': [
            'ProcessSubnets=subnetfinder.subnetfinder:ProcessSubnets'
        ]
    }
)



#######################################################
###  subnetfinder.py
###  This is the primary module.  This provides the 
###  ProcessSubnets method which really is all that 
###  should be called externally.
#######################################################

import ipaddress
import sys
import math

VERSION='1.2'
MIN_PYTHON = (3, 7)
if sys.version_info < MIN_PYTHON:
    sys.exit("Python %s.%s or later is required.\n" % MIN_PYTHON)

uniquelist=[]

def __findsupernets(sourcelist,showdebug):
    '''
    This method takes a list of ipaddress.IPv4Network objects and looks for supernets 
    which are completely contained by objects within the list, by stepping back up the
    CIDR tree one by one.  This module would not normally be called from another script.
    '''

    availablelist=[]
    count=0  # used as a counter for user information in debug
    #while more:
    more=answer=False
    # supernet=None

    sourcelist.sort()

    count+=1
    if len(sourcelist):
        prefix=sourcelist[0].prefixlen
        minprefix=int(prefix-math.log(len(sourcelist),2))
        if showdebug:
            print("Rollup scan pass %d.  Currently %d networks."%(count,len(sourcelist)))
            print("The calculated minimum supernet prefix lenth is %d"%(minprefix))
            print("Current available subnets")
            for x in sourcelist:
                print("%s"%(x))

    ###################################
    ###  This is the main loop which 
    ###  steps through each network
    ###  This selects one subnet and tries to find
    ###  the shortest prefix that can be represent
    ###  completely represented.
    ###################################
    
    for net in sourcelist:
        answer=True
        supernetprefix=prefix
        if showdebug:
            print("Checking %s"%(net))
        while answer & (supernetprefix>minprefix):
            supernetprefix-=1
            supernet=net.supernet(new_prefix=supernetprefix)
            if showdebug:
                print("Will try supernet %s"%(supernet))

            ###################################
            ###  This section could be optimized to make half
            ###  as many passes by recording which supernet's
            ###  subnets have already been checked.  Since this
            ###  entire script runs infrequently, it probably 
            ###  isn't worth the time.
            ###################################

            for x in supernet.subnets(new_prefix=prefix):
                answer &= (x in sourcelist)
                if showdebug & (x in sourcelist): # & (x != net):
                    print("Found subnet %s in available list"%(x))
                if x not in sourcelist:
                    if showdebug:
                        print("Didn't find %s in list"%(x))
                    break
            
            ###################################
            ###  If all subnets from the supernet
            ###  are found within the available list
            ###  then remove all subnets
            ###  and add the supernet.  Also set
            ###  the flag so to rescan the entire 
            ###  available list.
            ###################################

        if showdebug:
            print("Analysis shows supernet prefix %d, minimum prefix %d, answer %d"%(supernetprefix,minprefix,answer))    

        if (supernetprefix==minprefix) & answer:
            if showdebug:
                print("Special case:  Entire supernet is available.")
            return [ net.supernet(new_prefix=minprefix) ]
        elif supernetprefix+1<prefix:
            supernet=net.supernet(new_prefix=supernetprefix+1)
            if showdebug:
                print("Summarizing to /%s"%(supernet))
            for x in supernet.subnets(new_prefix=prefix):
                if showdebug:
                    print("Removing %s"%(x))
                sourcelist.remove(x)
            if supernet not in availablelist:
                if showdebug:
                    print("Adding %s"%(supernet))
                availablelist.append(supernet)
        else:
            if net not in availablelist:
                if showdebug:
                    print("Adding %s"%(supernet))
            availablelist.append(net)

    more |= answer #  set "more" to true if any supernets are found

    ###################################
    ###  If the loop found any supernets
    ###  the "while" loop continues
    ###################################
    
    if showdebug:
        print("Continue flags set to %s"%(more))

    return availablelist

def ProcessSubnets(supernet,usedlist,targetprefixlen,showdebug=False,screenout=False,summarize=False,uniquein=0):
    '''
    This will build a list of "seen" (used) subnets from the source provided, and then
    return a list of unseen subnets with length "prefixlen" from the same supernet.  
    It will ignore all provided subnets which are outside of the supernet range.  
    It will error and ignore if the line isn't a valid CIDR subnet (10.0.0.0/8) notation.  

    NOTE: The default CIDR length is omitted, a /32 is assumed.

    [Arguments]
    supernet:   The ipaddress.ip_network object for the supernet
    usedlist:   A iterable list of string objects.  Each string object must contain either a used 
                subnet in CIDR format, or a group header.  If a group header is supplied, subsequent 
                CIDR networks will be placed in that group.
                Example CIDR notation:  10.0.0.0/8
                Example group header:  Group: New Group
    targetprefixlen:  
                An int representing the target prefix length
    uniquein:   Can be used to find a subnet which exist only in N or less groups.
    showdebug:  Will print additional debugging information to STDOUT.
                Default is False.
    screenout:  Directs the output to STDOUT and exits rather than returning a list
                Default is False.
    supernets:  Combine contiguous blocks into summary supernets with shorter prefix length pl if possible.
                This may take more time if the list of available subnets is long.
    
    [Returns]
    This method returns a list of ipaddress.IPv4Network objects representing the unused subnets
    matching the characteristics provided.
    '''
    global uniquelist
    usedsubnets={}
    count=0
    if screenout:
        print("Beginning Read")

    ###################################
    ###  This is the main loop
    ###################################

    group='global'
    usedsubnets.update({group:{}})
    for line in usedlist:
        line=line.strip()
        count+=1
        if showdebug:
            print("Line %d:  Read '%s'"%(count,line))
        if line.lower().startswith('group:'):
            group=line[6:].strip()
            if group not in usedsubnets:
                usedsubnets.update({group:{}})
            if showdebug:
                print("line %d:  Changing to group '%s'"%(count,group))
            continue
        try:
            lineip=ipaddress.ip_network(line)
        except:
            if showdebug:
                print("line %d:  An error occured trying to convert '%s' into an IP subnet format"%(count,line))
            continue
        if showdebug:
            print("Line %d:  Converted '%s' to an IP object"%(count,str(lineip)))

        ###################################
        ###  Now, compare whether the used
        ###  subnet's prefix is longer,
        ###  shorter, or the same as the
        ###  target subnet.
        ###################################

        if (lineip.subnet_of(supernet)):
            if(lineip.prefixlen>targetprefixlen):  

                ## used prefix is longer, so add the target prefix length
                
                addsubnets={str(lineip.supernet(new_prefix=targetprefixlen)):1}
                usedsubnets[group].update(addsubnets)
                if showdebug:
                    print("Line %d:  Added %s which covers %s"%(count,list(addsubnets.keys()),str(lineip)))

            elif(lineip.prefixlen,targetprefixlen):

                ## used prefix is shorter, so add all the included target prefix lengths

                addsubnets={str(x):1 for x in lineip.subnets(new_prefix=targetprefixlen)}
                usedsubnets[group].update(addsubnets)
                if showdebug:
                    print("Line %d:  Added %s to include all of %s"%(count,list(addsubnets.keys()),lineip))

            else:

                ## used prefix is the same, so add just the used subnet

                usedsubnets[group].update({str(lineip):1})
                if showdebug:
                    print("Line %d:  Added %s"%(lineip))

        else:

            ### This line is only reached if the used subnet isn't within the supernet

            if showdebug:
                print("Line %d:  Ignoring '%s'.  Not within the supernet '%s'."%(count,line,str(supernet)))

    if screenout:
        print("Done Reading.  Read %d lines"%(count))

    ###################################
    ###  This generator object produces a complete list
    ###  of available subnets based on the supernet,
    ###  the target subnet length, the "unique in" value,
    ###  and the used subnets.
    ###################################    

    if showdebug:
        print("Full list of used subnets are as follows:")
        for x in usedsubnets.keys():
            print('Group: %s'%(x))
            for y in usedsubnets[x].keys():
                print(y)

    availablelist= [ x for x in supernet.subnets(new_prefix=targetprefixlen) 
        if str(x) not in [ z for y in usedsubnets.keys() 
        for z in usedsubnets[y].keys()] ]
    
    uniquelist= [ x for x in supernet.subnets(new_prefix=targetprefixlen) 
        if 0 <  [ z for y in usedsubnets.keys() 
        for z in usedsubnets[y].keys()].count(str(x)) <= uniquein ]

    if showdebug:
        print("Available Subnets")
        print(availablelist)
        for x in availablelist:
            print(x)
        print("Subnets matching uniqueness value")
        print(uniquelist)
        for x in uniquelist:
            print(x)

    availabledict={ x:1 for x in availablelist+uniquelist }
    availablelist=[x for x in availabledict.keys() ]
    availablelist.sort()

    if showdebug:
        print(availabledict)
        print("Combined Lists")
        for x in availablelist:
            print(x)

    if summarize:
        if screenout:
            print("Starting Rollup")
        availablelist=__findsupernets(availablelist,showdebug)

    return availablelist


